#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdint.h>
#include <linux/videodev2.h>
#include "../include/addWatermark.h"
#include "../include/NvLogging.h"

#define MAX(a, b) (a > b? a : b)
#define MIN(a, b) (a < b? a : b)


/****************************************************************************************************
* Function:     draw_font_func_yuv
* Description:  实现在yuv420图片上面画字 
* Input:        WATERMARK_PARAM param         
* Return:       
****************************************************************************************************/
int draw_font_func_yuv(WATERMARK_PARAM fParam)
{
    if (NULL == fParam.ptr_frame)
    {
        LOGE("input error!\n");
        return -1;
    }    
    if (NULL == fParam.font)
    {
        LOGE("imgOutput NULL!\n");
        return -1;
    }

    int tagY=0,tagU=0,tagV=0;
    uint8_t *offsetY=NULL,*offsetU=NULL,*offsetV=NULL;
    u_int8_t p8, mask8; // 一次读取8个dots
    
    /*yuv 地址的设置 */
    offsetY = fParam.ptr_frame;
    offsetU = offsetY + fParam.width * fParam.height;
    offsetV = offsetU + fParam.width * fParam.height/4;
    switch (fParam.color)
    {
        case 0:         // Yellow
            tagY = 226;tagU = 0;tagV = 149;
            break;
        case 1:         // Red
            tagY = 76;tagU = 85;tagV = 255;
            break;
        case 2:         // Green
            tagY = 150;tagU = 44;tagV = 21;
            break;
        case 3:         // Blue
            tagY = 29;tagU = 255;tagV = 107;
            break;
        case 4:         // White
            tagY = 255;tagU = 128;tagV = 128;
            break;
        default:        // grey
            tagY = 128;tagU = 128;tagV = 128;
            break;
    }  
    
    int x=0,y=0;
    int i=0, j=0;

    for (i = 0, y = fParam.starty; i < fParam.hFont && y < fParam.height - 1; i++, y++) //纵向偏移
    {
        p8 = *(u_int8_t *)(fParam.font + i*(fParam.wFont/8));  //首先从第i行取出字模的第一个字节，即前8个dot
        mask8 = 0x80;  /* 二进制 1000 0000 */
        for (j = 0, x = fParam.startx; j < fParam.wFont && x < fParam.width - 1; j++, x++) //横向偏移
        {
            if (p8 & mask8)
            {
                *(offsetY + y*fParam.width + x) = tagY;
                *(offsetU + (y/2) * (fParam.width/2) + x/2) = tagU;
                *(offsetV + (y/2) * (fParam.width/2) + x/2) = tagV;
            }
            mask8 = mask8 >> 1;  /* 循环移位取数据 */
            if (mask8 == 0)
            {
                p8 = *(u_int8_t *)(fParam.font + i*(fParam.wFont/8) + (j+1)/8);  //再从第i行取出第n个字节，
                mask8 = 0x80;
            }
        }
    }

    return 0;
}


/****************************************************************************************************
* Function:     draw_font_func_nv12
* Description:  实现在nv12图片上面画字 
* Input:        WATERMARK_PARAM param         
* Return:       
****************************************************************************************************/
int draw_font_func_nv12(WATERMARK_PARAM fParam)
{
    if (NULL == fParam.ptr_frame)
    {
        LOGE("input error!\n");
        return -1;
    }    
    if (NULL == fParam.font)
    {
        LOGE("imgOutput NULL!\n");
        return -1;
    }

    int tagY = 0, tagU = 0, tagV = 0;
    uint8_t *offsetY = NULL, *offsetUV = NULL;
    u_int8_t p8, mask8; // 一次读取8个dots
    
    /*yuv 地址的设置 */
    offsetY = fParam.ptr_frame;
    offsetUV = offsetY + fParam.width * fParam.height;

    switch (fParam.color)
    {
        case 0:         // Yellow
            tagY = 226; tagU = 0; tagV = 149;
            break;
        case 1:         // Red
            tagY = 76; tagU = 85; tagV = 255;
            break;
        case 2:         // Green
            tagY = 150; tagU = 44; tagV = 21;
            break;
        case 3:         // Blue
            tagY = 29; tagU = 255; tagV = 107;
            break;
        case 4:         // White
            tagY = 255; tagU = 128; tagV = 128;
            break;
        default:        // grey
            tagY = 128; tagU = 128; tagV = 128;
            break;
    }  
    
    int x = 0, y = 0;
    int i = 0, j = 0;
    int mid_val_4_uv = 0;
    for (i = 0, y = fParam.starty; i < fParam.hFont && y < fParam.height - 1; i++, y++) //纵向偏移
    {
        p8 = *(uint8_t *)(fParam.font + i * (fParam.wFont >> 3));  //首先从第i行取出字模的第一个字节，即前8个dot
        mask8 = 0x80;  /* 二进制 1000 0000 */
        for (j = 0, x = fParam.startx; j < fParam.wFont && x < fParam.width - 1; j++, x++) //横向偏移
        {
            if (p8 & mask8)
            {
                *(offsetY + y * fParam.width + x) = tagY;
                if (0 == (y & 0x01) && 0 == (x & 0x01))
                {
                    mid_val_4_uv = (y * fParam.width >> 1) + x;
                    *(offsetUV + mid_val_4_uv) = tagU;
                    *(offsetUV + mid_val_4_uv + 1) = tagV;
                }
            }
            mask8 = mask8 >> 1;  /* 循环移位取数据 */
            if (mask8 == 0)
            {
                p8 = *(uint8_t *)(fParam.font + i * (fParam.wFont >> 3) + ((j + 1) >> 3));  //再从第i行取出第n个字节，
                mask8 = 0x80;
            }
        }
    }

    return 0;
}


/****************************************************************************************************
* Function:     draw_font_func_yuyv
* Description:  实现在yuyv422图片上面画字 
* Input:        WATERMARK_PARAM param         
* Return:       
****************************************************************************************************/
int draw_font_func_yuyv(WATERMARK_PARAM fParam)
{
    if (NULL == fParam.ptr_frame)
    {
        LOGE("input error!\n");
        return -1;
    }    
    if (NULL == fParam.font)
    {
        LOGE("imgOutput NULL!\n");
        return -1;
    }

    int tagY=0, tagU=0, tagV=0;
    uint8_t *offsetY = NULL;
    u_int8_t p8, mask8; // 一次读取8个dots
    
    /*yuv 地址的设置 */
    offsetY = fParam.ptr_frame;
    switch (fParam.color)
    {
        case 0:         // Yellow
            tagY = 226;tagU = 0;tagV = 149;
            break;
        case 1:         // Red
            tagY = 76;tagU = 85;tagV = 255;
            break;
        case 2:         // Green
            tagY = 150;tagU = 44;tagV = 21;
            break;
        case 3:         // Blue
            tagY = 29;tagU = 255;tagV = 107;
            break;
        case 4:         // White
            tagY = 255;tagU = 128;tagV = 128;
            break;
        default:        // grey
            tagY = 128;tagU = 128;tagV = 128;
            break;
    }  
    
    int x=0,y=0;
    int i=0, j=0;
    int flag_u = 1;

    for (i = 0, y = fParam.starty; i < fParam.hFont && y < fParam.height - 1; i++, y++) //纵向偏移
    {
        p8 = *(u_int8_t *)(fParam.font + i*(fParam.wFont/8));  //首先从第i行取出字模的第一个字节，即前8个dot
        mask8 = 0x80;  /* 二进制 1000 0000 */
        uint8_t *tmp_ptr = offsetY + y * 2 * fParam.width + 2 * fParam.startx;
        for (j = 0, x = fParam.startx; j < fParam.wFont && x < fParam.width - 1; j++, x++) //横向偏移
        {
            if (p8 & mask8)
            {
                *(tmp_ptr + 2 * j) = tagY;
                if (flag_u)
                {
                    *(tmp_ptr + 2 * j + 1) = tagU;
                    flag_u = 0;
                }
                else
                {
                    *(tmp_ptr + 2 * j + 1) = tagV;
                    flag_u = 1;
                }
            }
            mask8 = mask8 >> 1;  /* 循环移位取数据 */
            if (mask8 == 0)
            {
                p8 = *(u_int8_t *)(fParam.font + i*(fParam.wFont/8) + (j+1)/8);  //再从第i行取出第n个字节，
                mask8 = 0x80;
            }
        }
    }

    return 0;
}

static int (*draw_Font_Func)(WATERMARK_PARAM fParam);

/****************************************************************************************************
* Function:     time_watermark
* Description:  添加时间水印 
* Input:        WATERMARK_PARAM tparam         
* Return:       
****************************************************************************************************/
int time_watermark(TIME_WATERM_PARAM tParam)
{
    char ftable[14][MAX_BYTE_ZIMO] = {0};
    u_int8_t spaceOffset = 0;
    u_int8_t i = 0;
    WATERMARK_PARAM fParam;
    time_t mtime;
    struct tm *ltime;
    struct timeval tm;

    fParam.ptr_frame = tParam.ptr_frame;
    fParam.width = tParam.width;
    fParam.height = tParam.height;
    fParam.color = tParam.color;
    if (16 != tParam.fontSize && 32 != tParam.fontSize && 64 != tParam.fontSize)
    {
        LOGE("fontSize input error!\n");
        return -1;
    }
    switch (tParam.fontSize)
    {
        case 16:
            for (i=0; i<14; i++)
            {
                memcpy(&ftable[i], &table_16[i], 16);
            }            
            spaceOffset = 8;
            fParam.wFont = 8;
            fParam.hFont = 16;
            break;
        case 32:
            for (i=0; i<14; i++)
            {
                memcpy(&ftable[i], &table_32[i], 64);
            }
            spaceOffset = 16;
            fParam.wFont = 16;
            fParam.hFont = 32;
            break;
        case 64:
            for (i=0; i<14; i++)
            {
                memcpy(&ftable[i], &table_64[i], 256);
            }
            spaceOffset = 32;
            fParam.wFont = 32;
            fParam.hFont = 64;
            break;
        default:
            break;
    }
    fParam.startx = tParam.startx;
    fParam.starty = tParam.starty;

    time(&mtime);
    ltime = localtime(&mtime);
    gettimeofday(&tm, NULL);
    if (V4L2_PIX_FMT_YUV420M == tParam.fmt)
    {
        draw_Font_Func = draw_font_func_yuv;
    }
    else if (V4L2_PIX_FMT_NV12 == tParam.fmt)
    {
        draw_Font_Func = draw_font_func_nv12;
    }
    else
    {
        draw_Font_Func = draw_font_func_yuyv;
    }

    for (i = 0; i < 23; i++)
    {
        fParam.startx += spaceOffset;
        switch (i)
        {
            case 0: //年第1位
                fParam.font = &ftable[((ltime->tm_year+1900)/1000)%10][0];
                draw_Font_Func(fParam);
                break;
            case 1: //年第2位
                fParam.font = &ftable[((ltime->tm_year+1900)/100)%10][0];
                draw_Font_Func(fParam);
                break;
            case 2: //年第3位
                fParam.font = &ftable[((ltime->tm_year+1900)/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 3: //年第4位
                fParam.font = &ftable[(ltime->tm_year+1900)%10][0];
                draw_Font_Func(fParam);
                break;
            case 4:
                fParam.font = &ftable[10][0];   //这里是连接符-
                draw_Font_Func(fParam);
                break;
            case 5: //月第1位
                fParam.font = &ftable[((ltime->tm_mon + 1)/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 6: //月第2位
                fParam.font = &ftable[(ltime->tm_mon + 1)%10][0];
                draw_Font_Func(fParam);
                break;
            case 7:
                fParam.font = &ftable[10][0];   //这里是连接符-
                draw_Font_Func(fParam);
                break;
            case 8: //日第1位
                fParam.font = &ftable[(ltime->tm_mday/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 9: //日第2位
                fParam.font = &ftable[(ltime->tm_mday)%10][0];
                draw_Font_Func(fParam);
                break;
            case 10:
                fParam.font = &ftable[12][0];   //这里是连接符 空格
                draw_Font_Func(fParam);
                break;
            case 11: //时第1位
                fParam.font = &ftable[(ltime->tm_hour/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 12: //时第2位
                fParam.font = &ftable[(ltime->tm_hour)%10][0];
                draw_Font_Func(fParam);
                break;
            case 13:
                fParam.font = &ftable[11][0];   //这里是连接符:
                draw_Font_Func(fParam);
                break;
            case 14: //分第1位
                fParam.font = &ftable[(ltime->tm_min/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 15: //分第2位
                fParam.font = &ftable[(ltime->tm_min)%10][0];
                draw_Font_Func(fParam);
                break;
            case 16:
                fParam.font = &ftable[11][0];   //这里是连接符:
                draw_Font_Func(fParam);
                break;
            case 17: //秒第1位
                fParam.font = &ftable[(ltime->tm_sec/10)%10][0];
                draw_Font_Func(fParam);
                break;
            case 18: //秒第2位
                fParam.font = &ftable[(ltime->tm_sec)%10][0];
                draw_Font_Func(fParam);
                break;
            case 19:
                fParam.font = &ftable[13][0];   //这里是连接符.
                draw_Font_Func(fParam);
                break;
            case 20: //毫秒第1位
                fParam.font = &ftable[(tm.tv_usec)/1000/100][0];
                draw_Font_Func(fParam);
                break;
            case 21: //毫秒第2位
                fParam.font = &ftable[(tm.tv_usec)/1000/10%10][0];
                draw_Font_Func(fParam);
                break;
            case 22: //毫秒第3位
                fParam.font = &ftable[(tm.tv_usec)/1000%10][0];
                draw_Font_Func(fParam);
                break;
            default:
                break;
        }
    }

    return 0;
}


/****************************************************************************************************
* Function:     direction_watermark
* Description:  添加方向水印 
* Input:        WATERMARK_PARAM tparam
                u_int8_t flag 1:前  2:后  3:左  4:右
* Return:       
****************************************************************************************************/
int direction_watermark(TIME_WATERM_PARAM tParam, u_int8_t flag)
{
    char ftable[4][MAX_BYTE_ZIMO] = {0};
    u_int8_t i = 0;
    WATERMARK_PARAM fParam;

    fParam.ptr_frame = tParam.ptr_frame;
    fParam.width = tParam.width;
    fParam.height = tParam.height;
    fParam.color = tParam.color;

    if (16 != tParam.fontSize && 32 != tParam.fontSize && 64 != tParam.fontSize)
    {
        LOGE("fontSize input error!\n");
        return -1;
    }
    if (1 != flag && 2 != flag && 3 != flag && 4 != flag)
    {
        LOGE("flag input error!\n");
        return -1;
    }
    switch (tParam.fontSize)
    {
        case 16:
            for (i=0; i<4; i++)
            {
                memcpy(&ftable[i], &table_direction16[i], 32);
            }
            fParam.wFont = 16;
            fParam.hFont = 16;
            break;
        case 32:
            for (i=0; i<4; i++)
            {
                memcpy(&ftable[i], &table_direction32[i], 128);
            }
            fParam.wFont = 32;
            fParam.hFont = 32;
            break;
        case 64:
            for (i=0; i<4; i++)
            {
                memcpy(&ftable[i], &table_direction64[i], 512);
            }
            fParam.wFont = 64;
            fParam.hFont = 64;
            break;
        default:
            break;
    }
    fParam.startx = tParam.startx;
    fParam.starty = tParam.starty;

    switch (flag)
    {
        case 1:
            fParam.font = &ftable[0][0];
            break;
        case 2:
            fParam.font = &ftable[1][0];
            break;
        case 3:
            fParam.font = &ftable[2][0];
            break;
        case 4:
            fParam.font = &ftable[3][0];
            break;
        default:
            break;
    }

    draw_Font_Func(fParam);
    return 0;
}