#include "./include/NvUtils.h"
#include <iostream>
#include <linux/videodev2.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>
#include <time.h>
#include <sys/prctl.h>
#include "nvenc_main.h"
#include "./include/addWatermark.h"


typedef struct _produce_yuv_arg_
{
    uint8_t *yuv_buf;
    pthread_mutex_t *read_camera;
    sem_t *enc_sem4file;
}produce_yuv_arg;

void *produce_yuv(void *arg)
{
    prctl(PR_SET_NAME, __FUNCTION__);
    produce_yuv_arg *yuv_arg = (produce_yuv_arg *)arg;

    TIME_WATERM_PARAM tParam;
    /*设置水印参数*/
    tParam.ptr_frame = yuv_arg->yuv_buf;
    tParam.width = 640;
    tParam.height = 480;
    tParam.fontSize = 32;
    tParam.color = 4;
    tParam.startx = 20;
    tParam.starty = 200;


    int ret = -1;
    /**> main loop */
    while (1)
    {

        if (0 != (ret = pthread_mutex_lock(yuv_arg->read_camera)))
        {
            LOGE("pthread_mutex_lock errnor!, errno=%d\n", ret);
        }
        memset(tParam.ptr_frame, 0, 640 * 480 * 3);
        tParam.fmt = V4L2_PIX_FMT_YUV420M;
        time_watermark(tParam);   //添加时间水印

        if (0 != (ret = pthread_mutex_unlock(yuv_arg->read_camera)))
        {
            LOGE("pthread_mutex_unlock errnor!, errno=%d\n", ret);
        }
        if (NULL != yuv_arg->enc_sem4file)
        {
            sem_post(yuv_arg->enc_sem4file);
        }
        usleep(65000);
    }

    return NULL;
}

/***********************************************
 * function: write h264 bitstream to save file.
 * input: src: h264 source bitstream
 *        len: one package length
 * return: 0 on succeed, -1 on failed. 
 * ********************************************/
int bitstream_write(unsigned char *src, uint32_t len)
{
    time_t timep;
    struct tm *ltime;
    static uint32_t file_size = 0;
    static char file_name[256] = {0};
    static FILE *fp = NULL;
    if (0 == len || len > 3145728)
    {
        LOGE("bitstreamLen illegal! len: %d\n", len);
        return 0;
    }

    if (0 == file_size)
    {
        time(&timep);
        ltime = localtime(&timep);
        memset(file_name, 0, sizeof(file_name));
        sprintf(file_name, "%04d-%02d-%02d_%02d-%02d-%02d.h264", ltime->tm_year + 1900, \
            ltime->tm_mon + 1, ltime->tm_mday, ltime->tm_hour, ltime->tm_min, ltime->tm_sec);

        fp = fopen(file_name, "ab+");
        if (NULL == fp)
        {
            LOGE("open %s faild, errno=%d\n", file_name, errno);
        }
    }

    if (NULL != fp)
    {
        LOGI("write file len: %d\n", len);
        fwrite(src, 1, len, fp);
        file_size += len;
    }

    if (file_size > 1024 * 1024 * 100)
    {
        file_size = 0;
        if (NULL != fp)
        {
            fclose(fp);
            fp = NULL;
        }
    }

    return 0;
}

int g_bitrate = 200;
void *get_video_bitrate(void *)
{
    int sleep_time = 0;
    while (1)
    {
        if (200 == g_bitrate)
        {
            g_bitrate = 500;
        }
        else if (500 == g_bitrate)
        {
            g_bitrate = 1000;
        }
        else
        {
            g_bitrate = 200;
        }
        srand((unsigned)time(NULL));
        sleep_time = rand() % 999 + 1;
        usleep(sleep_time * 1000);
    }
    return NULL;
}
/*********************************
 * function: creat a thread
 * input: func: callback function
 *        arg: the argument of callback function
 *        if_detach: if set the thread detached
 * output: tid: the thread id
 * return: 0 on success, ERRNO on faild.
 * *******************************/
int thread_create(void *func(void *), void *arg, uint8_t if_detach, pthread_t *tid)
{
    pthread_t id;
    int ret = -1;
    pthread_attr_t attr;
    ret = pthread_attr_init(&attr);
    if(0 != ret)
    {
        LOGE("pthread_attr_init faild!\n");
        return ret;
    }
    if (1 == if_detach)
    {
        ret = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        if(0 != ret)
        {
            LOGE("pthread_attr_setdetachstate faild!\n");
            return ret;
        }
    }
    ret = pthread_create(&id, &attr, func, arg);
    if(0 != ret)
    {
        LOGE("pthread_create faild!\n");
    }
    if (NULL != tid)
    {
        *tid = id;
    }
    return ret;
}

/*********************************
 * function: judgem if the thread is alive
 * input: the thread id
 * return: 0 is not alive, 1 is alive.
 * (attention: the thread should be joinable!)
 * *******************************/
int is_thread_alive(pthread_t tid)
{
    int b_alive = 0;
    if (0 != tid)
    {
        int ret = pthread_tryjoin_np(tid, NULL);
        switch (ret)
        {
            case 0:
                LOGI("thread has completed.\n");
                break;
            case EBUSY:
                b_alive = 1;
                break;
            case EINVAL:
                LOGE("thread is not a joinable thread.\n");
                break;
            case ESRCH:
                LOGE("no thread could be found with ID thread\n");
                break;
            case ENOENT:
                LOGE("thread couldn't find it's own valid handle\n");
                break;
            case EDEADLK:
                LOGE("attempt to join thread with self\n");
                break;
            default:
                LOGE("return unknown value, ret: %d\n", ret);
                break;
        }
    }
    return b_alive;
}

pthread_mutex_t read_yuv;	// lock the yuv buf
sem_t enc_sem4file; /**> this semaphore for the writing  file encoder. */

int main(int argc, char *argv[])
{
    prctl(PR_SET_NAME, __FUNCTION__);
    int ret = 0;
    pthread_t hw_enc_tid = 0;
    uint32_t l_bitrate = 200;
    uint8_t *pYuvBuf = (uint8_t *)malloc(640 * 480 * 3);

    produce_yuv_arg c_arg;
    THREAD_PARAM nvidia_param;
    if (0 != (ret = pthread_mutex_init(&read_yuv, NULL)))
    {
        LOGE("pthread_mutex_init error! return: %d\n", ret);
    }
    if (0 != sem_init(&enc_sem4file, 0, 0))
    {
        LOGE("sem_init error! errno=%d\n", errno);
        goto EXIT;
    }

    c_arg.yuv_buf = pYuvBuf;
    c_arg.read_camera = &read_yuv;
    c_arg.enc_sem4file = &enc_sem4file;
    if (0 != thread_create(get_video_bitrate, NULL, 1, NULL))
    {
        LOGE("thread_create get_video_bitrate error! errno=%d\n", errno);
        goto EXIT;
    }
    if (0 != thread_create(produce_yuv, (void *)&c_arg, 1, NULL))
    {
        LOGE("thread_create camera2yuv error! errno=%d\n", errno);
        goto EXIT;
    }

    while (1)
    {
        if (l_bitrate != g_bitrate)
        {
            l_bitrate = g_bitrate;
        }
        
        if (0 == is_thread_alive(hw_enc_tid))   //encoding with NVENC
        {
            memset(&nvidia_param, 0, sizeof(THREAD_PARAM));
            nvidia_param.fileSrc = pYuvBuf;
            nvidia_param.width = 640;
            nvidia_param.heigth = 480;
            nvidia_param.fps = 15;
            nvidia_param.encType = 1;
            nvidia_param.bitrate = l_bitrate * 1024;
            nvidia_param.read_camera = &read_yuv;
            nvidia_param.enc_sem = &enc_sem4file;
            nvidia_param.funcWriteBuf = bitstream_write;
            LOGI("fps=%d, l_bitrate=%d\n", nvidia_param.fps, l_bitrate * 1024);
            if (0 == thread_create(nvenc_main, (void *)&nvidia_param, 0, &hw_enc_tid))
            {
                LOGD("nvenc_main thread creat succeed!\n");
            }
        }
        usleep(1000);

    }

EXIT:

    if (-1 == sem_destroy(&enc_sem4file))
    {
        LOGE("sem_destroy error! errno=%d\n", errno);
    }
    free(pYuvBuf);

    return 0;
}